import typing as t
from pydantic import BaseModel


class TypeAndConfig(BaseModel):
    type: str
    config: t.Union[dict, str]


class StoreConfig(BaseModel):
    data_type: str
    type: str
    config: dict



from ._types import PresetData

variable_defaults = {
    "local_documents_path": "./documents",
    "encoding_model_name": "dunzhang/stella_en_1.5B_v5",
    "expected_hidden_size": 1536,
    "n_nearest": 10
}


_component_config = {
    "document_sources": {
        "local_directory": {
            "type": "filesystem",
            "config": {
                "root_path": "${local_documents_path}"
            }
        }
    },
    "storage": {
        "localpy": {
            "data_type": "omni",
            "type": "pydict",
            "config": {
                "dirpath": "./data",
                "overwrite": True
            }
        }
    },
    "llms": {
        "openai": {
            "type": "openai",
            "config": {
                "api_keyenvvar": "OPENAI_API_KEY"
            }
        }
    },
    "summarizers": {
        "pycode": {
            "type": "python_code",
            "config": {
                "llm": "openai"
            }
        }
    },
    "encoders": {
        "plaintext": {
            "type": "hugging_face",
            "config": {
                "model_name": "${encoding_model_name}",
                "tokenizer_config": {
                    "return_tensors": "pt",
                    "max_length": 1024,
                    "truncation": True,
                    "padding": "max_length"
                },
                "save_model": False,
                "expected_hidden_size": "${expected_hidden_size}"
            }
        }
    },
    "rag_agents": {
        "generic": {
            "type": "generic",
            "config": {
                "llm": "openai",
                "storage": "localpy",
                "encoder": "plaintext",
                "n_nearest": "${n_nearest}",
                "prompt": "",
                "system_prompt": ""
            }
        }
    }
}

_pipelines_config = {
    "ingest-directory-text": [
        {
            "action": "encode",
            "config": {
                "document_source": "local_directory",
                "encoder": "plaintext",
                "storage": "localpy"
            }
        }
    ]
}

_rag_query_command_config = {
    "rag_agent": "generic",
    "document_source": "local_directory"
}


local_docs_preset = PresetData(
    components=_component_config,
    piplines=_pipelines_config,
    rag_query_command=_rag_query_command_config,
    variable_defaults=variable_defaults
)



import json
from .bases import MetadataStore
from ..code_analysis.metadata_units.bases import ModuleData
from elasticsearch import Elasticsearch


class ElasticsearchMetadataStore(MetadataStore):

    name = 'elasticsearch'

    def __init__(self, config):
        self.index_name = config.pop('index_name')
        self._module_name_field = config.pop('module_name_field', "name")
        self._index_mapping_loc = config.pop('index_mapping_loc')
        self.config = config
        self.es = Elasticsearch([self.config])
    
    def store_all_module_data(self, modules: dict[str, ModuleData]):
        for _, module_data in modules.items():
            self.store_module_data(module_data)

    def store_module_data(self, module_data: ModuleData):
        doc = module_data.model_dump()

    def _create_index(self):
        with open(self._index_mapping_loc) as f:
            index_mapping = json.load(f)
        self.es.indices.create(index=self.index_name, body=index_mapping)

    def query_modules(self, query):
        return self.es.search(index=self.index_name, body=query)

    def get_module(self, module_name):
        query = {
            'query': {
                'term': {
                    self._module_name_field: module_name
                }
            }
        }
        response = self.es.search(index=self.index_name, body=query)
        return response['hits']['hits'][0]['_source']
    


import os
from pydantic import BaseModel, Field
from .bases import DocumentSourceBase
from ..storage.store_factory import get_store_cls
from ..storage.bases import TextDocumentStore, VectorStore
from ..common_types import StoreConfig



class TextStoreDocumentSource(DocumentSourceBase):
    
    name = "storage"

    def __init__(self, config: StoreConfig):
        super().__init__(config)
        self.config = config
        self._text_doc_store: TextDocumentStore = self._initialize_text_doc_store()
    
    def _initialize_text_doc_store(self):
        store_cls = get_store_cls(self.config.data_type, self.config.type)
        store_config = self.config.config
        return store_cls(store_config)

    def get_documents(self, document_names: list[str] = None) -> dict[str, str]:
        if document_names:
            return {
                document_name: self._text_doc_store.get_document(document_name)
                for document_name in document_names
            }
        return self._text_doc_store.get_all_documents()



from logging import getLogger
import os
from pathlib import Path
import typing as t

import yaml
from deepmerge import always_merger, Merger

from ._types import MasterConfig
from ...actions.bases import ActionConfig
from ...actions.encode import EncodeActionConfig
from ...actions.summarize import SummarizeActionConfig
from ...actions.rag import RagActionConfig
from ...rag.bases import TypeAndConfig, StoreConfig
from ..configuration.presets.preset_factory import get_preset, PresetData

logger = getLogger(__name__)


def load_config(configpath: Path = None) -> MasterConfig:
    with open(str(configpath)) as f:
        config = yaml.safe_load(f)
    return MasterConfig(**config)


def get_preset_config(preset_name, **vars) -> MasterConfig:
    preset = get_preset(preset_name)
    return preset.get_config(**vars)


def merge_defaults(config: MasterConfig,
                   preset_data: PresetData = None,
                   **vars
                   ) -> MasterConfig:
    config_d = config.model_dump()
    component_config = config_d.get("components", {})
    pipelines_config = config_d.get("pipelines", {})
    rag_query_command = config_d.get("rag_query_command", {})
    
    component_config = always_merger.merge(
        preset_data.get_component_config(**vars),
        component_config
    )

    pipelines_merger = Merger(
        [
            (dict, "merge"),
            (list, "override")
        ],
        ["override"],
        ["override"]
    )
    pipelines_config = pipelines_merger.merge(
        preset_data.get_pipelines_config(**vars),
        pipelines_config
    )
    rag_query_command = always_merger.merge(
        preset_data.get_rag_query_command_config(**vars),
        rag_query_command
    )
    return MasterConfig(
        project_name=config.project_name,
        components=component_config,
        pipelines=pipelines_config,
        rag_query_command=rag_query_command
    )

class ActionConfigFactory:
    
    def __init__(self, master_config: MasterConfig) -> None:
        self.master_config = master_config

    def dereference_action_config(self, action_config: ActionConfig) -> ActionConfig:
        raise NotImplementedError

    def dereference_document_source(self, document_source: TypeAndConfig) -> TypeAndConfig:

        if isinstance(document_source, str):
            document_source = self.master_config.components.document_sources[document_source]

        if all([
            document_source.type == "storage",
            isinstance(document_source.config, str)
        ]):
            source_name = document_source.config
            source_config = self.master_config.components.storage[source_name]
            document_source.config = source_config
        return document_source

    def dereference_storage(self, storage: t.Union[str, StoreConfig]) -> StoreConfig:
        if isinstance(storage, str):
            return self.master_config.components.storage[storage]
        return storage

    def dereference_llm(self, llm: t.Union[str, TypeAndConfig]) -> TypeAndConfig:
        if isinstance(llm, str):
            return self.master_config.components.llms[llm]
        return llm

    def dereference_encoder(self, encoder: t.Union[str, TypeAndConfig]) -> TypeAndConfig:
        if isinstance(encoder, str):
            return self.master_config.components.encoders[encoder]
        return encoder

    def dereference_summarizer(self, summarizer: t.Union[str, TypeAndConfig]) -> TypeAndConfig:
        if isinstance(summarizer, str):
            return self.master_config.components.summarizers[summarizer]
        return summarizer


class EncodeActionConfigFactory(ActionConfigFactory):

    def dereference_action_config(self, action_config: EncodeActionConfig) -> EncodeActionConfig:
        action_config.encoder = self.dereference_encoder(action_config.encoder)
        action_config.document_source = self.dereference_document_source(action_config.document_source)
        action_config.storage = self.dereference_storage(action_config.storage)        
        return action_config
    

class SummarizeActionConfigFactory(ActionConfigFactory):

    def dereference_action_config(self, action_config: SummarizeActionConfig) -> SummarizeActionConfig:
        action_config.summarizer = self.dereference_summarizer(action_config.summarizer)
        action_config.storage = self.dereference_storage(action_config.storage)
        action_config.summarizer.config.llm =\
            self.dereference_llm(action_config.summarizer.config.llm) 
        action_config.document_source = self.dereference_document_source(action_config.document_source)
        return action_config


class RagActionConfigFactory(ActionConfigFactory):
    
    def dereference_action_config(self, action_config: RagActionConfig) -> RagActionConfig:
        action_config.rag_agent.config.llm = self.dereference_llm(action_config.rag_agent.config.llm)
        action_config.rag_agent.config.storage = self.dereference_storage(action_config.rag_agent.config.storage)
        action_config.rag_agent.config.encoder = self.dereference_encoder(action_config.rag_agent.config.encoder)
        action_config.document_source = self.dereference_document_source(action_config.document_source)
        return action_config


def get_action_config_factory(action_name: str, master_config: MasterConfig) -> ActionConfigFactory:
    _actions = {
        "encode": EncodeActionConfigFactory,
        "summarize": SummarizeActionConfigFactory,
        "rag": RagActionConfigFactory
    }
    if action_name not in _actions:
        raise ValueError(f"Action '{action_name}' not found.")
    return _actions[action_name](master_config)
